import re
from typing import Set, Iterable, Tuple


class ScopeError(ValueError):
    pass


class ScopeValidationError(ScopeError):
    pass


class ScopeAuthorizationError(ScopeError):
    pass


FIXED_SCOPES = {
    "av",
    "av:ro",
    "av:hosting",
    "av:hosting:ro",
    "av:mailbox",
    "av:services",
    "av:services:ro",
}

FIXED_REGEXES = [
    re.compile(r"^av:hosting:agent1[qpzry9x8gf2tvdw0s3jn54khce6mua7l]{59}$"),
    re.compile(r"^av:mailbox:agent1[qpzry9x8gf2tvdw0s3jn54khce6mua7l]{59}$"),
    re.compile(
        r"^av:services:[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}$"
    ),
]


def validate_scope(value: str):
    # check the fixed patterns
    if value in FIXED_SCOPES:
        return

    # check each of the regexes
    for pattern in FIXED_REGEXES:
        if pattern.match(value) is not None:
            return

    raise ScopeValidationError(f"Invalid scope: {value}")


def validate_scopes(values: Iterable[str]):
    for value in values:
        validate_scope(value)


def parse_scopes(value: str) -> Set[str]:
    tokens = set(value.split())
    validate_scopes(tokens)
    return tokens


def normalise_scopes(value: str | Set[str]) -> str:
    if isinstance(value, str):
        scopes = parse_scopes(value)
    else:
        scopes = value

    return " ".join(sorted(scopes))


def _extract_read_only(value: str) -> Tuple[str, bool]:
    read_only = False
    if value.endswith(":ro"):
        value = value[:-3]
        read_only = True
    return value, read_only


def _authorize_scope(scope: str, requirement: str) -> bool:

    # strip and remove the read only flags from the requirement
    scope, scope_ro = _extract_read_only(scope)
    requirement, requirement_ro = _extract_read_only(requirement)

    # if the requirement is NOT read only, but the scope
    if scope_ro and not requirement_ro:
        return False

    # split the scope into a tree
    requirement_tree = requirement.split(":")
    scope_tree = scope.split(":")

    # if the scope has LOWER permissions than the requirement tree, then this
    # is an automatic auth failure
    if len(scope_tree) > len(requirement_tree):
        return False

    common_tree_len = min(len(scope_tree), len(requirement_tree))
    authorized = scope_tree[:common_tree_len] == requirement_tree[:common_tree_len]

    return authorized


def authorize_scope(
    scope: str, requirement: str, *, raise_exception: bool = True
) -> bool:
    success = _authorize_scope(scope, requirement)
    if not success and raise_exception:
        raise ScopeAuthorizationError()

    return success


def authorize_scopes(scopes: str | Set[str], requirement: str) -> bool:
    if isinstance(scopes, str):
        scopes = parse_scopes(scopes)

    success = False
    for scope in scopes:
        if authorize_scope(scope, requirement, raise_exception=False):
            success = True
            break

    if not success:
        raise ScopeAuthorizationError()

    return True



from enum import Enum
from typing import Optional, Set

from pydantic import BaseModel


class ClientType(str, Enum):
    User = "user"  # An active user session
    Delegate = "delegate"  # a delegated user session
    App = "app"  # an app session


class Client(BaseModel):
    client_type: ClientType
    client_id: str
    group: Optional[str] = None
    scopes: Set[str]



from typing import Dict, Optional

from fastapi import HTTPException, status

from .types import Client


class TestingTokenRegister:
    def __init__(self):
        self.tokens: Dict[str, Client] = {}

    def get(self, token: str) -> Optional[Client]:
        return self.tokens.get(token)

    def set(self, token: str, client: Client):
        self.tokens[token] = client

    def delete(self, token: str):
        del self.tokens[token]

    def clear(self):
        self.tokens.clear()


test_token_register = TestingTokenRegister()


def mocked_auth(token: str) -> Client:
    client = test_token_register.get(token)
    if client is None:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
        )

    return client



import json
import logging
import re
from typing import Optional

from fastapi import Request, Depends, HTTPException, status
from fastapi.security import OAuth2AuthorizationCodeBearer
from jwcrypto import jwt, jwk

from .public_key import FAUNA_PUBLIC_KEY
from .scopes import authorize_scopes, ScopeAuthorizationError, parse_scopes
from .testing import mocked_auth
from .types import Client, ClientType

USER_ID_PATTERN = re.compile(r"^[a-z0-9]{48}$")

public_key = jwk.JWK.from_pem(FAUNA_PUBLIC_KEY.encode())


class FaunaAuthorizationScheme(OAuth2AuthorizationCodeBearer):
    async def __call__(self, request: Request) -> Optional[str]:
        fauna = request.cookies.get("fauna")
        if fauna is not None:
            return fauna

        return await super().__call__(request)


fauna_auth_scheme = FaunaAuthorizationScheme(authorizationUrl="", tokenUrl="token")


def _parse_token(token: str) -> Optional[Client]:
    try:
        jwt_token = jwt.JWT(
            key=public_key,
            jwt=token,
            check_claims={"exp": None, "iss": "fetch.ai", "sub": None, "iat": None},
        )

        # parse the claims
        claims = json.loads(jwt_token.claims)

        # extract the elements
        subject = str(claims["sub"])
        group = claims.get("grp")

        # ensure that we match the user id pattern
        if USER_ID_PATTERN.match(subject) is None:
            return None

        # capture the scopes, if not present then generate an empty set
        scope = claims.get("scope")
        if scope is not None:
            scopes = parse_scopes(scope)
        else:
            scopes = set()

        # determine client type by the presence of the scopes
        if len(scopes) == 0:
            client_type = ClientType.User
        else:
            client_type = ClientType.Delegate

        return Client(
            client_type=client_type,
            client_id=subject,
            group=group,
            scopes=scopes,
        )

    except jwt.JWException:
        return None


def _is_authorization_mocked(request: Request) -> bool:
    return request.app.state._state.get("authorization_mocked", False)


def get_fauna_client(
    request: Request, token: Optional[str] = Depends(fauna_auth_scheme)
) -> Client:
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
    )

    if token is None:
        raise credentials_exception

    if _is_authorization_mocked(request):
        logging.warning("Using mocked authorization")
        return mocked_auth(token)

    # normal authorization flow
    client = _parse_token(token)
    if client is None:
        raise credentials_exception

    return client


def get_fauna_user(client: Client = Depends(get_fauna_client)) -> Client:
    if client.client_type != ClientType.User:
        raise HTTPException(status_code=401, detail="Invalid token")
    return client


class GetFaunaClient:
    def __init__(self, requirement_pattern: Optional[str] = None):
        if requirement_pattern is not None and requirement_pattern.endswith(":ro"):
            raise ValueError(
                "Do not specify the read only permission via the requirement. This will be computed automatically by method"
            )

        self._pattern = requirement_pattern

    def __call__(
        self, request: Request, client: Client = Depends(get_fauna_client)
    ) -> Client:
        credentials_exception = HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
        )

        # simple path - the authorized client is of user type
        if client.client_type == ClientType.User:
            return client

        # sanity check
        assert client.client_type != ClientType.User

        # no delegated permission pattern - should result in an error
        if self._pattern is None:
            raise credentials_exception

        # sanity check
        assert self._pattern is not None

        # resolve the scope permissions
        try:
            requirement = self._pattern.format(**request.path_params)
            if request.method.lower() == "get":
                requirement += ":ro"

        except KeyError:
            raise credentials_exception

        # validate the scopes that are required
        try:
            authorize_scopes(client.scopes, requirement)
        except ScopeAuthorizationError:
            raise credentials_exception

        # all checks are passed - return the client
        return client
